# -*- coding: utf-8 -*-

"""
The :mod:`audiotsm.io.base` module provides base classes for the input and
output of :class:`~audiotsm.base.tsm.TSM` objects.
"""


class Reader(object):
    """An abstract class for the input of a :class:`~audiotsm.base.tsm.TSM`
    object."""

    @property
    def channels(self):
        """The number of channels of the :class:`~audiotsm.io.base.Reader`."""
        raise NotImplementedError()

    @property
    def empty(self):
        """True if there is no more data to read."""
        raise NotImplementedError()

    def read(self, buffer):
        """Reads as many samples from the :class:`~audiotsm.io.base.Reader` as
        possible, write them to ``buffer``, and returns the number of samples
        that were read.

        :param buffer: a matrix of shape (``m``, ``n``), with ``m`` the number
            of channels and ``n`` the length of the buffer, where the samples
            will be written.
        :type buffer: :class:`numpy.ndarray`
        :returns: the number of samples that were read. It should always be
            equal to the length of the buffer, except when there is no more
            values to be read.
        :raises ValueError: if the :class:`~audiotsm.io.base.Reader` and the
            buffer do not have the same number of channels
        """
        raise NotImplementedError()

    def skip(self, n):
        """Try to skip ``n`` samples, an returns the number of samples that
        were actually skipped."""
        raise NotImplementedError()


class Writer(object):
    """An abstract class for the output of a :class:`~audiotsm.base.tsm.TSM`
    object."""

    @property
    def channels(self):
        """The number of channels of the :class:`~audiotsm.io.base.Writer`."""
        raise NotImplementedError()

    def write(self, buffer):
        """Write as many samples from the :class:`~audiotsm.io.base.Writer` as
        possible from ``buffer``, and returns the number of samples that were
        written.

        :param buffer: a matrix of shape (``m``, ``n``), with ``m`` the number
            of channels and ``n`` the length of the buffer, where the samples
            will be read.
        :type buffer: :class:`numpy.ndarray`
        :returns: the number of samples that were written. It should always be
            equal to the length of the buffer, except when there is no more
            space in the :class:`~audiotsm.io.base.Writer`.
        :raises ValueError: if the :class:`~audiotsm.io.base.Writer` and the
            buffer do not have the same number of channels
        """
        raise NotImplementedError()



# -*- coding: utf-8 -*-

"""
The :mod:`audiotsm.io.stream` module provides a
:class:`~audiotsm.io.base.Writer` allowing to play the output of a
:class:`~audiotsm.base.tsm.TSM` object in real-time.
"""


import numpy as np
from sounddevice import OutputStream

from . import base


class StreamWriter(base.Writer):
    """A :class:`~audiotsm.io.base.Writer` allowing to play the output of a
    :class:`~audiotsm.base.tsm.TSM` object directly.

    You should stop the :class:`~audiotsm.io.stream.StreamWriter` after using
    it with the :func:`~audiotsm.io.stream.StreamWriter.stop` method, or use it
    in a ``with`` statement as follow::

        with WavWriter(2, 44100) as writer:
            # use writer...

    :param channels: the number of channels of the signal.
    :type channels: int
    :param samplerate: the sampling rate of the signal.
    :type samplerate: int
    :param attrs: additional parameters used to create the
        :class:`sounddevice.OutputStream` that is used by the
        :class:`~audiotsm.io.stream.StreamWriter`.
    """
    def __init__(self, channels, samplerate, **attrs):
        self._channels = channels

        self._stream = OutputStream(samplerate=samplerate, channels=channels,
                                    **attrs)
        self._stream.start()

    @property
    def channels(self):
        return self._channels

    def write(self, buffer):
        if buffer.shape[0] != self.channels:
            raise ValueError("the buffer should have the same number of "
                             "channels as the WavWriter")

        self._stream.write(np.ascontiguousarray(buffer.T))

        return buffer.shape[1]

    def stop(self):
        """Stop the stream."""
        self._stream.stop()

    def __enter__(self):
        return self

    def __exit__(self, _1, _2, _3):
        self.stop()



# -*- coding: utf-8 -*-

"""
Tests for the audiotsm.io.array package.
"""

import pytest
import numpy as np
from numpy.testing import assert_almost_equal

from audiotsm.io.array import ArrayReader, ArrayWriter, FixedArrayWriter


@pytest.mark.parametrize("data_in, read_out, n_out, data_out", [
    ([[]], [[]], 0, [[]]),
    ([[]], [[0]], 0, [[]]),
    ([[1, 2, 3], [4, 5, 6]], [[], []], 0, [[1, 2, 3], [4, 5, 6]]),
    ([[1, 2, 3], [4, 5, 6]], [[1], [4]], 1, [[2, 3], [5, 6]]),
    ([[1, 2, 3], [4, 5, 6]], [[1, 2], [4, 5]], 2, [[3], [6]]),
    ([[1, 2, 3], [4, 5, 6]], [[1, 2, 3], [4, 5, 6]], 3, [[], []]),
    ([[1, 2, 3], [4, 5, 6]], [[1, 2, 3, 0], [4, 5, 6, 0]], 3, [[], []]),
])
def test_read(data_in, read_out, n_out, data_out):
    """Run tests for the ArrayReader.read method."""
    reader = ArrayReader(np.array(data_in))

    buffer = np.zeros_like(read_out, dtype=np.float32)
    n = reader.read(buffer)
    assert_almost_equal(buffer, read_out)
    assert n == n_out

    # Check the data remaining in the reader
    buffer = np.zeros_like(data_out)
    reader.read(buffer)
    assert_almost_equal(buffer, data_out)

    # Check that there is no more data in the reader
    buffer = np.zeros_like(data_in)
    n = reader.read(buffer)
    assert not buffer.any()
    assert n == 0


@pytest.mark.parametrize("data_in, n_in, n_out, data_out", [
    ([[]], 0, 0, [[]]),
    ([[]], 1, 0, [[]]),
    ([[1, 2, 3], [4, 5, 6]], 0, 0, [[1, 2, 3], [4, 5, 6]]),
    ([[1, 2, 3], [4, 5, 6]], 1, 1, [[2, 3], [5, 6]]),
    ([[1, 2, 3], [4, 5, 6]], 2, 2, [[3], [6]]),
    ([[1, 2, 3], [4, 5, 6]], 3, 3, [[], []]),
    ([[1, 2, 3], [4, 5, 6]], 4, 3, [[], []]),
])
def test_skip(data_in, n_in, n_out, data_out):
    """Run tests for the ArrayReader.skip method."""
    reader = ArrayReader(np.array(data_in))

    n = reader.skip(n_in)
    assert n == n_out

    # Check the data remaining in the reader
    buffer = np.zeros_like(data_out)
    reader.read(buffer)
    assert_almost_equal(buffer, data_out)

    # Check that there is no more data in the reader
    buffer = np.zeros_like(data_in)
    n = reader.read(buffer)
    assert not buffer.any()
    assert n == 0


@pytest.mark.parametrize("write1, write2, n1_out, n2_out, buffer_out", [
    ([[], []], [[], []], 0, 0, [[], []]),

    ([[1, 2, 3], [4, 5, 6]], [[], []], 3, 0, [[1, 2, 3], [4, 5, 6]]),
    ([[1, 2], [4, 5]], [[3], [6]], 2, 1, [[1, 2, 3], [4, 5, 6]]),
    ([[1], [4]], [[2, 3], [5, 6]], 1, 2, [[1, 2, 3], [4, 5, 6]]),
    ([[], []], [[1, 2, 3], [4, 5, 6]], 0, 3, [[1, 2, 3], [4, 5, 6]]),

    ([[1, 2, 3], [4, 5, 6]], [[7], [8]], 3, 0, [[1, 2, 3], [4, 5, 6]]),
    ([[1, 2, 3], [4, 5, 6]], [[7], [8]], 2, 0, [[1, 2], [4, 5]]),

    ([[1, 2], [4, 5]], [[], []], 2, 0, [[1, 2, 0], [4, 5, 0]]),
])
def test_fixed_array_write(write1, write2, n1_out, n2_out, buffer_out):
    """Run tests for the FixedArrayWriter.write method."""
    buffer = np.zeros_like(buffer_out, dtype=np.float32)
    writer = FixedArrayWriter(buffer)

    n = writer.write(np.array(write1, dtype=np.float32))
    assert n == n1_out
    n = writer.write(np.array(write2, dtype=np.float32))
    assert n == n2_out

    assert_almost_equal(buffer, buffer_out)


@pytest.mark.parametrize("write1, write2, n1_out, n2_out, buffer_out", [
    ([[], []], [[], []], 0, 0, [[], []]),

    ([[1, 2, 3], [4, 5, 6]], [[], []], 3, 0, [[1, 2, 3], [4, 5, 6]]),
    ([[1, 2], [4, 5]], [[3], [6]], 2, 1, [[1, 2, 3], [4, 5, 6]]),
    ([[1], [4]], [[2, 3], [5, 6]], 1, 2, [[1, 2, 3], [4, 5, 6]]),
    ([[], []], [[1, 2, 3], [4, 5, 6]], 0, 3, [[1, 2, 3], [4, 5, 6]]),

    ([[1, 2], [4, 5]], [[], []], 2, 0, [[1, 2], [4, 5]]),
])
def test_array_write(write1, write2, n1_out, n2_out, buffer_out):
    """Run tests for the ArrayWriter.write method."""
    writer = ArrayWriter(len(write1))

    n = writer.write(np.array(write1, dtype=np.float32))
    assert n == n1_out
    n = writer.write(np.array(write2, dtype=np.float32))
    assert n == n2_out

    assert_almost_equal(writer.data, buffer_out)



# -*- coding: utf-8 -*-

"""
The :mod:`audiotsm.ola` module implements the OLA (Overlap-Add) time-scale
modification procedure.
"""

from audiotsm.base import AnalysisSynthesisTSM, Converter
from audiotsm.utils.windows import hanning


class OLAConverter(Converter):
    """A Converter implementing the OLA (Overlap-Add) time-scale modification
    procedure."""
    def convert_frame(self, analysis_frame):
        return analysis_frame


def ola(channels, speed=1., frame_length=256, analysis_hop=None,
        synthesis_hop=None):
    """Returns a :class:`~audiotsm.base.tsm.TSM` object implementing the OLA
    (Overlap-Add) time-scale modification procedure.

    In most cases, you should not need to set the ``frame_length``, the
    ``analysis_hop`` or the ``synthesis_hop``. If you want to fine tune these
    parameters, you can check the documentation of the
    :class:`~audiotsm.base.analysis_synthesis.AnalysisSynthesisTSM` class to
    see what they represent.

    :param channels: the number of channels of the input signal.
    :type channels: int
    :param speed: the speed ratio by which the speed of the signal will be
        multiplied (for example, if ``speed`` is set to 0.5, the output signal
        will be half as fast as the input signal).
    :type speed: float, optional
    :param frame_length: the length of the frames.
    :type frame_length: int, optional
    :param analysis_hop: the number of samples between two consecutive analysis
        frames (``speed * synthesis_hop`` by default). If ``analysis_hop`` is
        set, the ``speed`` parameter will be ignored.
    :type analysis_hop: int, optional
    :param synthesis_hop: the number of samples between two consecutive
        synthesis frames (``frame_length // 2`` by default).
    :type synthesis_hop: int, optional
    :returns: a :class:`audiotsm.base.tsm.TSM` object
    """
    if synthesis_hop is None:
        synthesis_hop = frame_length // 2

    if analysis_hop is None:
        analysis_hop = int(synthesis_hop * speed)

    analysis_window = None
    synthesis_window = hanning(frame_length)

    converter = OLAConverter()

    return AnalysisSynthesisTSM(
        converter, channels, frame_length, analysis_hop, synthesis_hop,
        analysis_window, synthesis_window)



# -*- coding: utf-8 -*-

"""
Tests for the audiotsm.utils.normalizebuffer package.
"""

import pytest
from numpy.testing import assert_almost_equal

from audiotsm.utils import NormalizeBuffer


def generate_normalize_buffers(array):
    """Generate different NormalizeBuffers containing the same data as
    ``array``."""
    for i in range(0, len(array)):
        buffer = NormalizeBuffer(len(array))
        buffer.remove(i)
        buffer.add(array)
        yield buffer


def generate_test_cases(cases):
    """For each tuple (data, *params) of an array, generate multiple test cases
    (buffer, *params), where buffer is a NormalizeBuffer containing
    ``data``."""
    for case in cases:
        for buffer in generate_normalize_buffers(case[0]):
            yield (buffer,) + case[1:]


@pytest.mark.parametrize("buffer, window, out", generate_test_cases([
    ([], [], []),
    ([0.5, 1.5, 2.5], [0.25, 0, -0.25], [0.75, 1.5, 2.25]),
    ([0.5, 1.5, 2.5, 3.5, 4.5], [0.25, 0, -0.25], [0.75, 1.5, 2.25, 3.5, 4.5])
]))
def test_normalize_buffer_add(buffer, window, out):
    """Run tests for the NormalizeBuffer.add method."""
    buffer.add(window)
    assert_almost_equal(buffer.to_array(), out)


@pytest.mark.parametrize("buffer, n, out", generate_test_cases([
    ([], 0, []),
    ([1], 0, [1]),
    ([1], 1, [0]),
    ([1, 0.75, 0.5], 0, [1, 0.75, 0.5]),
    ([1, 0.75, 0.5], 1, [0.75, 0.5, 0]),
    ([1, 0.75, 0.5], 2, [0.5, 0, 0]),
    ([1, 0.75, 0.5], 3, [0, 0, 0]),
    ([1, 0.75, 0.5], 4, [0, 0, 0]),
]))
def test_normalize_buffer_remove(buffer, n, out):
    """Run tests for the NormalizeBuffer.remove method."""
    buffer.remove(n)
    assert_almost_equal(buffer.to_array(), out)


@pytest.mark.parametrize("buffer, start, end, out", generate_test_cases([
    ([], 0, 0, []),
    ([1], 0, 0, []),
    ([1], 0, 1, [1]),
    ([1, 0.75, 0.5], 0, 0, []),
    ([1, 0.75, 0.5], 0, 1, [1]),
    ([1, 0.75, 0.5], 0, 2, [1, 0.75]),
    ([1, 0.75, 0.5], 0, 3, [1, 0.75, 0.5]),
    ([1, 0.75, 0.5], 1, 1, []),
    ([1, 0.75, 0.5], 1, 2, [0.75]),
    ([1, 0.75, 0.5], 1, 3, [0.75, 0.5]),
    ([1, 0.75, 0.5], 2, 2, []),
    ([1, 0.75, 0.5], 2, 3, [0.5]),
    ([1, 0.75, 0.5], 3, 3, []),
]))
def test_normalize_buffer_to_array(buffer, start, end, out):
    """Run tests for the NormalizeBuffer.to_array method."""
    assert_almost_equal(buffer.to_array(start, end), out)



# -*- coding: utf-8 -*-

"""
The :mod:`audiotsm.utils` module provides utility functions and classes used in
the implementation of time-scale modification procedures.

.. autoclass:: audiotsm.utils.CBuffer
    :members:
    :undoc-members:

.. autoclass:: audiotsm.utils.NormalizeBuffer
    :members:
    :undoc-members:
"""

from .cbuffer import CBuffer
from .normalizebuffer import NormalizeBuffer



# -*- coding: utf-8 -*-

"""
The :mod:`audiotsm.base.tsm` module provides an abstract class for real-time
audio time-scale modification procedures.
"""


class TSM(object):
    """An abstract class for real-time audio time-scale modification
    procedures.

    If you want to use a :class:`~audiotsm.base.tsm.TSM` object to run a TSM
    procedure on a signal, you should use the
    :func:`~audiotsm.base.tsm.TSM.run` method in most cases.
    """

    def clear(self):
        """Clears the state of the :class:`~audiotsm.base.tsm.TSM` object,
        making it ready to be used on another signal (or another part of a
        signal).

        This method should be called before processing a new file, or seeking
        to another part of a signal.
        """
        raise NotImplementedError

    def flush_to(self, writer):
        """Writes as many output samples as possible to ``writer``, assuming
        that there are no remaining samples that will be added to the input
        (i.e. that the :func:`~audiotsm.base.tsm.TSM.write_to` method will not
        be called), and returns the number of samples that were written.

        :param writer: a :class:`audiotsm.io.base.Writer`.
        :returns: a tuple (``n``, ``finished``), with:

            - ``n`` the number of samples that were written to ``writer``
            - ``finished`` a boolean that is ``True`` when there are no samples
              remaining to flush.
        :rtype: (int, bool)
        """
        raise NotImplementedError

    def read_from(self, reader):
        """Reads as many samples as possible from ``reader``, processes them,
        and returns the number of samples that were read.

        :param reader: a :class:`audiotsm.io.base.Reader`.
        :returns: the number of samples that were read from ``reader``.
        """
        raise NotImplementedError

    def run(self, reader, writer):
        """Runs the TSM procedure on the content of ``reader`` and writes the
        output to ``writer``.

        :param reader: a :class:`audiotsm.io.base.Reader`.
        :param writer: a :class:`audiotsm.io.base.Writer`.
        """
        finished = False
        while not (finished and reader.empty):
            self.read_from(reader)
            _, finished = self.write_to(writer)

        finished = False
        while not finished:
            _, finished = self.flush_to(writer)

        self.clear()

    def set_speed(self, speed):
        """Sets the speed ratio.

        :param speed: the speed ratio by which the speed of the signal will be
            multiplied (for example, if ``speed`` is set to 0.5, the output
            signal will be half as fast as the input signal).
        :type speed: float
        """
        raise NotImplementedError

    def write_to(self, writer):
        """Writes as many result samples as possible to ``writer``.

        :param writer: a :class:`audiotsm.io.base.Writer`.
        :returns: a tuple (``n``, ``finished``), with:

            - ``n`` the number of samples that were written to ``writer``
            - ``finished`` a boolean that is ``True`` when there are no samples
              remaining to write. In this case, the
              :func:`~audiotsm.base.tsm.TSM.read_from` method should be called
              to add new input samples, or, if there are no remaining input
              samples, the :func:`~audiotsm.base.tsm.TSM.flush_to` method
              should be called to get the last output samples.
        :rtype: (int, bool)
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-

"""
The :mod:`audiotsm.base.analysis_synthesis` module provides a base class for
real-time analysis-synthesis based audio time-scale modification procedures.
"""

import numpy as np

from audiotsm.utils import (windows, CBuffer, NormalizeBuffer)
from .tsm import TSM

EPSILON = 0.0001


class AnalysisSynthesisTSM(TSM):
    """A :class:`audiotsm.base.tsm.TSM` for real-time analysis-synthesis based
    time-scale modification procedures.

    The basic principle of an analysis-synthesis based TSM procedure is to
    first decompose the input signal into short overlapping frames, called the
    analysis frames. The frames have a fixed length ``frame_length``, and are
    separated by ``analysis_hop`` samples, as illustrated below::

                 <--------frame_length--------><-analysis_hop->
       Frame 1:  [~~~~~~~~~~~~~~~~~~~~~~~~~~~~]
       Frame 2:                  [~~~~~~~~~~~~~~~~~~~~~~~~~~~~]
       Frame 3:                                  [~~~~~~~~~~~~~~~~~~~~~~~~~~~~]
       ...

    It then relocates the frames on the time axis by changing the distance
    between them (to ``synthesis_hop``), as illustrated below::

                 <--------frame_length--------><----synthesis_hop---->
       Frame 1:  [~~~~~~~~~~~~~~~~~~~~~~~~~~~~]
       Frame 2:                         [~~~~~~~~~~~~~~~~~~~~~~~~~~~~]
       Frame 3:                                               [~~~~~~~~~~~~~~~~~~~~~~~~~~~~]
       ...

    This changes the speed of the signal by the ratio ``analysis_hop /
    synthesis_hop`` (for example, if the ``synthesis_hop`` is twice the
    ``analysis_hop``, the output signal will be half as fast as the input
    signal).

    However this simple method introduces artifacts to the signal. These
    artifacts can be reduced by modifying the analysis frames by various
    methods. This is done by a ``converter`` object, which converts the
    analysis frames into modified frames called the synthesis frames.

    To further reduce the artifacts, window functions (the ``analysis_window``
    and the ``synthesis_window``) can be applied to the analysis frames and the
    synthesis frames in order to smooth the signal.

    Some TSM procedures (e.g. WSOLA-like methods) may need to have access to
    some samples preceeding or following an analysis frame to generate the
    synthesis frame. The `delta_before` and `delta_after` parameters allow to
    specify the numbers of samples needed before and after the analysis frame,
    so that they are available to the ``converter``.

    For more details on Time-Scale Modification procedures, I recommend reading
    "`A Review of Time-Scale Modification of Music Signals`_" by Jonathan
    Driedger and Meinard Müller.

    .. _A Review of Time-Scale Modification of Music Signals:
        http://www.mdpi.com/2076-3417/6/2/57

    :param converter: an object that implements the conversion of the analysis
        frames into synthesis frames.
    :type converter: :class:`Converter`
    :param channels: the number of channels of the input signal.
    :type channels: int
    :param frame_length: the length of the frames.
    :type frame_length: int
    :param analysis_hop: the number of samples between two consecutive analysis
        frames.
    :type analysis_hop: int
    :param synthesis_hop: the number of samples between two consecutive
        synthesis frames.
    :type synthesis_hop: int
    :param analysis_window: a window applied to the analysis frames
    :type analysis_window: :class:`numpy.ndarray`
    :param synthesis_window: a window applied to the synthesis frames
    :type synthesis_window: :class:`numpy.ndarray`
    :param delta_before: the number of samples preceding an analysis frame that
        the converter requires (this is usually 0, except for WSOLA-like
        methods)
    :type delta_before: int
    :param delta_after: the number of samples following an analysis frame that
        the converter requires (this is usually 0, except for WSOLA-like
        methods)
    :type delta_after: int
    """  # noqa: E501
    # pylint: disable=too-many-instance-attributes
    def __init__(self, converter, channels, frame_length, analysis_hop,
                 synthesis_hop, analysis_window, synthesis_window,
                 delta_before=0, delta_after=0):
        # pylint: disable=too-many-arguments
        self._converter = converter

        self._channels = channels
        self._frame_length = frame_length
        self._analysis_hop = analysis_hop
        self._synthesis_hop = synthesis_hop

        self._analysis_window = analysis_window
        self._synthesis_window = synthesis_window

        self._delta_before = delta_before
        self._delta_after = delta_after

        # When the analysis hop is larger than the frame length, some samples
        # from the input need to be skipped. self._skip_input_samples tracks
        # how many samples should be skipped before reading the analysis frame.
        self._skip_input_samples = 0

        # This attribute is used to start the output signal in the middle of a
        # frame, which should be the peek of the window function
        self._skip_output_samples = 0

        # Compute the normalize window
        self._normalize_window = windows.product(self._analysis_window,
                                                 self._synthesis_window)

        if self._normalize_window is None:
            self._normalize_window = np.ones(self._frame_length)

        # Initialize the buffers
        delta = self._delta_before + self._delta_after
        self._in_buffer = CBuffer(self._channels, self._frame_length + delta)
        self._analysis_frame = np.empty(
            (self._channels, self._frame_length + delta))
        self._out_buffer = CBuffer(self._channels, self._frame_length)
        self._normalize_buffer = NormalizeBuffer(self._frame_length)

        self.clear()

    def clear(self):
        # Clear the buffers
        self._in_buffer.remove(self._in_buffer.length)
        self._out_buffer.remove(self._out_buffer.length)
        self._out_buffer.right_pad(self._frame_length)
        self._normalize_buffer.remove(self._normalize_buffer.length)

        # Left pad the input with half a frame of zeros, and ignore that half
        # frame in the output. This makes the output signal start in the middle
        # of a frame, which should be the peak of the window function.
        self._in_buffer.write(np.zeros(
            (self._channels, self._delta_before + self._frame_length // 2)))
        self._skip_output_samples = self._frame_length // 2

        # Clear the converter
        self._converter.clear()

    def flush_to(self, writer):
        if self._in_buffer.remaining_length == 0:
            raise RuntimeError("There is still data to process in the input "
                               "buffer, flush_to method should only be called "
                               "when write_to returns True.")

        n = self._out_buffer.write_to(writer)
        if self._out_buffer.ready == 0:
            # The output buffer is empty
            self.clear()
            return n, True

        return n, False

    def _process_frame(self):
        """Read an analysis frame from the input buffer, process it, and write
        the result to the output buffer."""
        # Generate the analysis frame and discard the input samples that will
        # not be needed anymore
        self._in_buffer.peek(self._analysis_frame)
        self._in_buffer.remove(self._analysis_hop)

        # Apply the analysis window
        windows.apply(self._analysis_frame, self._analysis_window)

        # Convert the analysis frame into a synthesis frame
        synthesis_frame = self._converter.convert_frame(self._analysis_frame)

        # Apply the synthesis window
        windows.apply(synthesis_frame, self._synthesis_window)

        # Overlap and add the synthesis frame in the output buffer
        self._out_buffer.add(synthesis_frame)

        # The overlap and add step changes the volume of the signal. The
        # normalize_buffer is used to keep track of "how much of the input
        # signal was added" to each part of the output buffer, allowing to
        # normalize it.
        self._normalize_buffer.add(self._normalize_window)

        # Normalize the samples that are ready to be written to the output
        normalize = self._normalize_buffer.to_array(end=self._synthesis_hop)
        normalize[normalize < EPSILON] = 1
        self._out_buffer.divide(normalize)
        self._out_buffer.set_ready(self._synthesis_hop)
        self._normalize_buffer.remove(self._synthesis_hop)

    def read_from(self, reader):
        n = reader.skip(self._skip_input_samples)
        self._skip_input_samples -= n
        if self._skip_input_samples > 0:
            return n

        n += self._in_buffer.read_from(reader)

        if (self._in_buffer.remaining_length == 0 and
                self._out_buffer.remaining_length >= self._synthesis_hop):
            # The input buffer has enough data to process, and there is enough
            # space in the output buffer to store the output
            self._process_frame()

            # Skip output samples if necessary
            skipped = self._out_buffer.remove(self._skip_output_samples)
            self._out_buffer.right_pad(skipped)
            self._skip_output_samples -= skipped

            # Set the number of input samples to be skipped
            self._skip_input_samples = self._analysis_hop - self._frame_length
            if self._skip_input_samples < 0:
                self._skip_input_samples = 0

        return n

    def set_speed(self, speed):
        self._analysis_hop = int(self._synthesis_hop * speed)

    def write_to(self, writer):
        n = self._out_buffer.write_to(writer)
        self._out_buffer.right_pad(n)

        if (self._in_buffer.remaining_length > 0 and
                self._out_buffer.ready == 0):
            # There is not enough data to process in the input buffer, and the
            # output buffer is empty
            return n, True

        return n, False


class Converter(object):
    """A base class for objects implementing the conversion of analysis frames
    into synthesis frames."""

    def clear(self):
        """Clears the state of the Converter, making it ready to be used on
        another signal (or another part of a signal). It is called by the
        :func:`~audiotsm.base.tsm.TSM.clear` method and the constructor of
        :class:`AnalysisSynthesisTSM`."""
        # pylint: disable=no-self-use
        return

    def convert_frame(self, analysis_frame):
        """Converts an analysis frame into a synthesis frame.

        :param analysis_frame: a matrix of shape (``m``, ``delta_before +
            frame_length + delta_after``), with ``m`` the number of channels,
            containing the analysis frame and some samples before and after
            (as specified by the ``delta_before`` and ``delta_after``
            parameters of the :class:`AnalysisSynthesisTSM` calling the
            :class:`Converter`).

            ``analysis_frame[:, delta_before:-delta_after]`` contains the
            actual analysis frame (without the samples preceeding and following
            it).
        :type analysis_frame: :class:`numpy.ndarray`
        :returns: a synthesis frame represented as a :class:`numpy.ndarray` of
            shape (``m``, ``frame_length``), with ``m`` the number of channels.
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-

"""
The :mod:`audiotsm.utils.windows` module contains window functions used for
digital signal processing.
"""

import numpy as np


def apply(buffer, window):
    """Applies a window to a buffer.

    :param buffer: a matrix of shape (``m``, ``n``), with ``m`` the number of
        channels and ``n`` the length of the buffer.
    :type buffer: :class:`numpy.ndarray`
    :param window: a :class:`numpy.ndarray` of shape (``n``,).
    """
    if window is None:
        return

    for channel in buffer:
        channel *= window


def hanning(length):
    """Returns a periodic Hanning window.

    Contrary to :func:`numpy.hanning`, which returns the symetric Hanning
    window, :func:`hanning` returns a periodic Hanning window, which is better
    for spectral analysis.

    :param length: the number of points of the Hanning window
    :type length: :class:`int`
    :return: the window as a :class:`numpy.ndarray` of shape (``length``,).
    """
    if length <= 0:
        return np.zeros(0)

    time = np.arange(length)
    return 0.5 * (1 - np.cos(2 * np.pi * time / length))


def product(window1, window2):
    """Returns the product of two windows.

    :param window1: a :class:`numpy.ndarray` of shape (``n``,) or ``None``.
    :param window2: a :class:`numpy.ndarray` of shape (``n``,) or ``None``.
    :returns: the product of the two windows. If one of the windows is equal to
        ``None``, the other is returned, and if the two are equal to ``None``,
        ``None`` is returned.
    """
    if window1 is None:
        return window2

    if window2 is None:
        return window1

    return window1 * window2



# -*- coding: utf-8 -*-

"""
The :mod:`audiotsm.utils.cbuffer` module implements a circular buffer used to
store multichannel audio data.
"""

import numpy as np


class CBuffer(object):
    """A :class:`CBuffer` is a circular buffer used to store multichannel audio
    data.

    It can be seen as a variable-size buffer whose length is bounded by
    ``max_length``. The :func:`CBuffer.write` and :func:`CBuffer.right_pad`
    methods allow to add samples at the end of the buffer, while the
    :func:`CBuffer.read` and :func:`CBuffer.remove` methods allow to remove
    samples from the beginning of the buffer.

    Contrary to the samples added by the :func:`CBuffer.write` and
    :func:`CBuffer.read_from`, those added by the :func:`CBuffer.right_pad`
    method are considered not to be ready to be read. Effectively, this means
    that they can be modified by the :func:`CBuffer.add` and
    :func:`CBuffer.divide` methods, but have to be marked as ready to be read
    with the :func:`CBuffer.set_ready` method before being read with the
    :func:`CBuffer.peek`, :func:`CBuffer.read`, or :func:`CBuffer.write_to`
    methods.

    :param channels: the number of channels of the buffer.
    :type channels: int
    :param max_length: the maximum length of the buffer (i.e. the maximum
        number of samples that can be stored in each channel).
    :type max_length: int
    """
    def __init__(self, channels, max_length):
        self._data = np.zeros((channels, max_length), dtype=np.float32)
        self._channels = channels
        self._max_length = max_length

        self._offset = 0
        self._ready = 0
        self._length = 0

    def __repr__(self):
        return "CBuffer(offset={}, length={}, ready={}, data=\n{})".format(
            self._offset, self._length, self._ready, repr(self.to_array()))

    def add(self, buffer):
        """Adds a ``buffer`` element-wise to the :class:`CBuffer`.

        :param buffer: a matrix of shape (``m``, ``n``), with ``m`` the number
            of channels and ``n`` the length of the buffer.
        :type buffer: :class:`numpy.ndarray`
        :raises ValueError: if the :class:`CBuffer` and the ``buffer`` do not
            have the same number of channels or the :class:`CBuffer` is smaller
            than the ``buffer`` (``self.length < n``).
        """
        if buffer.shape[0] != self._data.shape[0]:
            raise ValueError("the two buffers should have the same number of "
                             "channels")

        n = buffer.shape[1]
        if n > self._length:
            raise ValueError("not enough space remaining in CBuffer")

        # Compute the slice of data where the values will be added
        start = self._offset
        end = self._offset + n

        if end <= self._max_length:
            self._data[:, start:end] += buffer[:, :n]
        else:
            end -= self._max_length
            self._data[:, start:] += buffer[:, :self._max_length - start]
            self._data[:, :end] += buffer[:, self._max_length - start:n]

    def divide(self, array):
        """Divides each channel of the :class:`CBuffer` element-wise by the
        ``array``.

        :param array: an array of shape (``n``,).
        :type array: :class:`numpy.ndarray`
        :raises ValueError: if the length of the :class:`CBuffer` is smaller
            than the length of the array (``self.length < n``).
        """
        n = len(array)
        if n > self._length:
            raise ValueError("not enough space remaining in the CBuffer")

        # Compute the slice of data where the values will be divided
        start = self._offset
        end = self._offset + n

        if end <= self._max_length:
            self._data[:, start:end] /= array[:n]
        else:
            end -= self._max_length
            self._data[:, start:] /= array[:self._max_length - start]
            self._data[:, :end] /= array[self._max_length - start:n]

    @property
    def length(self):
        """The number of samples of each channel of the :class:`CBuffer`."""
        return self._length

    def peek(self, buffer):
        """Reads as many samples from the :class:`CBuffer` as possible, without
        removing them from the :class:`CBuffer`, writes them to the ``buffer``,
        and returns the number of samples that were read.

        The samples need to be marked as ready to be read with the
        :func:`CBuffer.set_ready` method in order to be read. This is done
        automatically by the :func:`CBuffer.write` and
        :func:`CBuffer.read_from` methods.

        :param buffer: a matrix of shape (``m``, ``n``), with ``m`` the number
            of channels and ``n`` the length of the buffer, where the samples
            will be written.
        :type buffer: :class:`numpy.ndarray`
        :returns: the number of samples that were read from the
            :class:`CBuffer`.
        :raises ValueError: if the :class:`CBuffer` and the ``buffer`` do not
            have the same number of channels.
        """
        if buffer.shape[0] != self._data.shape[0]:
            raise ValueError("the two buffers should have the same number of "
                             "channels")

        n = min(buffer.shape[1], self._ready)

        # Compute the slice of data the values will be read from
        start = self._offset
        end = self._offset + n

        if end <= self._max_length:
            np.copyto(buffer[:, :n], self._data[:, start:end])
        else:
            end -= self._max_length
            np.copyto(buffer[:, :self._max_length - start],
                      self._data[:, start:])
            np.copyto(buffer[:, self._max_length - start:n],
                      self._data[:, :end])

        return n

    def read(self, buffer):
        """Reads as many samples from the :class:`CBuffer` as possible, removes
        them from the :class:`CBuffer`, writes them to the ``buffer``, and
        returns the number of samples that were read.

        The samples need to be marked as ready to be read with the
        :func:`CBuffer.set_ready` method in order to be read. This is done
        automatically by the :func:`CBuffer.write` and
        :func:`CBuffer.read_from` methods.

        :param buffer: a matrix of shape (``m``, ``n``), with ``m`` the number
            of channels and ``n`` the length of the buffer, where the samples
            will be written.
        :type buffer: :class:`numpy.ndarray`
        :returns: the number of samples that were read from the
            :class:`CBuffer`.
        :raises ValueError: if the :class:`CBuffer` and the ``buffer`` do not
            have the same number of channels.
        """
        n = self.peek(buffer)
        self.remove(n)
        return n

    def read_from(self, reader):
        """Reads as many samples as possible from ``reader``, writes them to
        the :class:`CBuffer`, and returns the number of samples that were read.

        The written samples are marked as ready to be read.

        :param reader: a :class:`audiotsm.io.base.Reader`.
        :returns: the number of samples that were read from ``reader``.
        :raises ValueError: if the :class:`CBuffer` and ``reader`` do not have
            the same number of channels.
        """
        # Compute the slice of data that will be written to
        start = (self._offset + self._length) % self._max_length
        end = start + self._max_length - self._length

        if end <= self._max_length:
            n = reader.read(self._data[:, start:end])
        else:
            # There is not enough space to copy the whole buffer, it has to be
            # split into two parts, one of which will be copied at the end of
            # _data, and the other at the beginning.
            end -= self._max_length

            n = reader.read(self._data[:, start:])
            n += reader.read(self._data[:, :end])

        self._length += n
        self._ready = self._length
        return n

    @property
    def ready(self):
        """The number of samples that can be read."""
        return self._ready

    @property
    def remaining_length(self):
        """The number of samples that can be added to the :class:`CBuffer`."""
        return self._max_length - self._ready

    def remove(self, n):
        """Removes the first ``n`` samples of the :class:`CBuffer`, preventing
        them to be read again, and leaving more space for new samples to be
        written.

        :param n: the number of samples to remove.
        :type n: int
        :returns: the number of samples that were removed.
        """
        if n >= self._length:
            n = self._length

        # Compute the slice of data that will be reset to 0
        start = self._offset
        end = self._offset + n

        if end <= self._max_length:
            self._data[:, start:end] = 0
        else:
            end -= self._max_length
            self._data[:, start:] = 0
            self._data[:, :end] = 0

        self._offset += n
        self._offset %= self._max_length
        self._length -= n

        self._ready -= n
        if self._ready < 0:
            self._ready = 0

        return n

    def right_pad(self, n):
        """Add zeros at the end of the :class:`CBuffer`.

        The added samples are not marked as ready to be read. The
        :func:`CBuffer.set_ready` will need to be called in order to be able to
        read them.

        :param n: the number of zeros to add.
        :type n: int
        :raises ValueError: if there is not enough space to add the zeros.
        """
        if n > self._max_length - self._length:
            raise ValueError("not enough space remaining in :class:`CBuffer`")

        self._length += n

    def set_ready(self, n):
        """Mark the next ``n`` samples as ready to be read.

        :param n: the number of samples to mark as ready to be read.
        :type n: int
        :raises ValueError: if there is less than ``n`` samples that are not
            ready yet.
        """
        if self._ready + n > self._length:
            raise ValueError("not enough samples to be marked as ready")

        self._ready += n

    def to_array(self):
        """Returns an array containing the same data as the :class:`CBuffer`.

        :returns: a :class:`numpy.ndarray` of shape (``m``, ``n``), with ``m``
            the number of channels and ``n`` the length of the buffer.
        """
        out = np.empty((self._channels, self._ready))
        self.peek(out)
        return out

    def write(self, buffer):
        """Writes as many samples from the ``buffer`` to the :class:`CBuffer`
        as possible, and returns the number of samples that were read.

        The written samples are marked as ready to be read.

        :param buffer: a matrix of shape (``m``, ``n``), with ``m``
            the number of channels and ``n`` the length of the buffer, where
            the samples will be read.
        :type buffer: :class:`numpy.ndarray`
        :returns: the number of samples that were written to the
            :class:`CBuffer`.
        :raises ValueError: if the :class:`CBuffer` and the ``buffer`` do not
            have the same number of channels.
        """
        if buffer.shape[0] != self._data.shape[0]:
            raise ValueError("the two buffers should have the same number of "
                             "channels")

        n = min(buffer.shape[1], self._max_length - self._length)

        # Compute the slice of data that will be written to
        start = (self._offset + self._length) % self._max_length
        end = start + n

        if end <= self._max_length:
            np.copyto(self._data[:, start:end], buffer[:, :n])
        else:
            # There is not enough space to copy the whole buffer, it has to be
            # split into two parts, one of which will be copied at the end of
            # _data, and the other at the beginning.
            end -= self._max_length

            np.copyto(self._data[:, start:],
                      buffer[:, :self._max_length - start])
            np.copyto(self._data[:, :end],
                      buffer[:, self._max_length - start:n])

        self._length += n
        self._ready = self._length
        return n

    def write_to(self, writer):
        """Writes as many samples as possible to ``writer``, deletes them from
        the :class:`CBuffer`, and returns the number of samples that were
        written.

        The samples need to be marked as ready to be read with the
        :func:`CBuffer.set_ready` method in order to be read. This is done
        automatically by the :func:`CBuffer.write` and
        :func:`CBuffer.read_from` methods.

        :param writer: a :class:`audiotsm.io.base.Writer`.
        :returns: the number of samples that were written to ``writer``.
        :raises ValueError: if the :class:`CBuffer` and ``writer`` do not have
            the same number of channels.
        """
        # Compute the slice of data the values will be read from
        start = self._offset
        end = self._offset + self._ready

        if end <= self._max_length:
            n = writer.write(self._data[:, start:end])
        else:
            end -= self._max_length
            n = writer.write(self._data[:, start:])
            n += writer.write(self._data[:, :end])

        self.remove(n)
        return n



# -*- coding: utf-8 -*-

"""
The :mod:`audiotsm.io.wav` module provides a :class:`~audiotsm.io.base.Reader`
and a :class:`~audiotsm.io.base.Writer` allowing to use wav files as input or
output of a :class:`~audiotsm.base.tsm.TSM` object.
"""

import wave
import numpy as np

from . import base


class WavReader(base.Reader):
    """A :class:`~audiotsm.io.base.Reader` allowing to use a wav file as input
    of a :class:`~audiotsm.base.tsm.TSM` object.

    You should close the :class:`~audiotsm.io.wav.WavReader` after using it
    with the :func:`~audiotsm.io.wav.WavReader.close` method, or use it in a
    ``with`` statement as follow::

        with WavReader(filename) as reader:
            # use reader...

    :param filename: the name of an existing wav file.
    :type filename: str
    """
    def __init__(self, filename):
        self._reader = wave.open(filename, 'rb')

    @property
    def channels(self):
        return self._reader.getnchannels()

    @property
    def empty(self):
        return self._reader.tell() == self._reader.getnframes()

    def close(self):
        """Close the wav file."""
        self._reader.close()

    def read(self, buffer):
        if buffer.shape[0] != self.channels:
            raise ValueError("the buffer should have the same number of "
                             "channels as the WavReader")

        frames = self._reader.readframes(buffer.shape[1])
        frames = np.frombuffer(frames, '<i2').astype(np.float32) / 32676

        # Separate channels
        frames = frames.reshape((-1, self.channels)).T

        n = frames.shape[1]
        np.copyto(buffer[:, :n], frames)
        del frames

        return n

    @property
    def samplerate(self):
        """The samplerate of the wav file."""
        return self._reader.getframerate()

    @property
    def samplewidth(self):
        """The sample width in bytes of the wav file."""
        return self._reader.getsamplewidth()

    def skip(self, n):
        current_pos = self._reader.tell()
        new_pos = min(current_pos + n, self._reader.getnframes())

        self._reader.setpos(new_pos)

        return new_pos - current_pos

    def __enter__(self):
        return self

    def __exit__(self, _1, _2, _3):
        self.close()


class WavWriter(base.Writer):
    """A :class:`~audiotsm.io.base.Writer` allowing to use a wav file as output
    of a :class:`~audiotsm.base.tsm.TSM` object.

    You should close the :class:`~audiotsm.io.wav.WavWriter` after using it
    with the :func:`~audiotsm.io.wav.WavWriter.close` method, or use it in a
    ``with`` statement as follow::

        with WavWriter(filename, 2, 44100) as writer:
            # use writer...

    :param filename: the name of the wav file (it will be overwritten if it
        already exists).
    :type filename: str
    :param channels: the number of channels of the signal.
    :type channels: int
    :param samplerate: the sampling rate of the signal.
    :type samplerate: int
    """
    def __init__(self, filename, channels, samplerate):
        self._writer = wave.open(filename, 'wb')
        self._channels = channels
        self._writer.setnchannels(channels)
        self._writer.setframerate(samplerate)
        self._writer.setsampwidth(2)

    @property
    def channels(self):
        return self._channels

    def close(self):
        """Close the wav file."""
        self._writer.close()

    def write(self, buffer):
        if buffer.shape[0] != self.channels:
            raise ValueError("the buffer should have the same number of "
                             "channels as the WavWriter")

        n = buffer.shape[1]
        frames = (buffer.T.reshape((-1,)) * 32676).astype(np.int16).tobytes()
        self._writer.writeframes(frames)
        del frames

        return n

    def __enter__(self):
        return self

    def __exit__(self, _1, _2, _3):
        self.close()



# -*- coding: utf-8 -*-

"""
audiotsm.base
~~~~~~~~~~~~~

This module provides base classes for the implementation of time-scale
modification procedures.
"""

from .tsm import TSM
from .analysis_synthesis import AnalysisSynthesisTSM, Converter



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
sine
~~~~

Run a TSM procedure on a signal generated with numpy.
"""
# pylint: disable=invalid-name

import numpy as np
import sounddevice as sd
from audiotsm import wsola
from audiotsm.io.array import ArrayReader, ArrayWriter


# The parameters of the input signal
length = 1  # in seconds
samplerate = 44100  # in Hz
frequency = 440  # an A4

# Generate the input signal
time = np.linspace(0, length, int(length * samplerate))
input_signal = np.sin(np.pi * frequency * time).reshape((1, -1))

# Run the TSM procedure
reader = ArrayReader(input_signal)
writer = ArrayWriter(channels=1)

tsm = wsola(channels=1, speed=0.5)
tsm.run(reader, writer)

# Play the output
# This example was written to show how to use an ArrayWriter. If you want to
# play the output of a TSM procedure you should use an
# audiotsm.io.stream.StreamWriter.
sd.play(np.ascontiguousarray(writer.data.T), samplerate, blocking=True)



# -*- coding: utf-8 -*-

"""
The :mod:`audiotsm.wsola` module implements the WSOLA (Waveform
Similarity-based Overlap-Add) time-scale modification procedure.

WSOLA works in the same way as OLA, with the exception that it allows slight
shift of the position of the analysis frames.
"""

import numpy as np

from audiotsm.base import AnalysisSynthesisTSM, Converter
from audiotsm.utils.windows import hanning


class WSOLAConverter(Converter):
    """A Converter implementing the WSOLA (Waveform Similarity-based
    Overlap-Add) time-scale modification procedure."""
    def __init__(self, channels, frame_length, synthesis_hop, tolerance):
        self._channels = channels
        self._frame_length = frame_length
        self._synthesis_hop = synthesis_hop
        self._tolerance = tolerance

        self._synthesis_frame = np.empty((channels, frame_length))
        self._natural_progression = np.empty((channels, frame_length))
        self._first = True

    def clear(self):
        self._first = True

    def convert_frame(self, analysis_frame):
        for k in range(0, self._channels):
            if self._first:
                delta = 0
            else:
                cross_correlation = np.correlate(
                    analysis_frame[k, :-self._synthesis_hop],
                    self._natural_progression[k])
                delta = np.argmax(cross_correlation)
                del cross_correlation

            # Copy the shifted analysis frame to the synthesis frame buffer
            np.copyto(self._synthesis_frame[k],
                      analysis_frame[k, delta:delta + self._frame_length])

            # Save the natural progression (what the next synthesis frame would
            # be at normal speed)
            delta += self._synthesis_hop
            np.copyto(self._natural_progression[k],
                      analysis_frame[k, delta:delta + self._frame_length])

        self._first = False

        return self._synthesis_frame


def wsola(channels, speed=1., frame_length=1024, analysis_hop=None,
          synthesis_hop=None, tolerance=None):
    """Returns a :class:`~audiotsm.base.tsm.TSM` object implementing the WSOLA
    (Waveform Similarity-based Overlap-Add) time-scale modification procedure.

    In most cases, you should not need to set the ``frame_length``, the
    ``analysis_hop``, the ``synthesis_hop``, or the ``tolerance``. If you want
    to fine tune these parameters, you can check the documentation of the
    :class:`~audiotsm.base.analysis_synthesis.AnalysisSynthesisTSM` class to
    see what the first three represent.

    WSOLA works in the same way as OLA, with the exception that it allows
    slight shift (at most ``tolerance``) of the position of the analysis
    frames.

    :param channels: the number of channels of the input signal.
    :type channels: int
    :param speed: the speed ratio by which the speed of the signal will be
        multiplied (for example, if ``speed`` is set to 0.5, the output signal
        will be half as fast as the input signal).
    :type speed: float, optional
    :param frame_length: the length of the frames.
    :type frame_length: int, optional
    :param analysis_hop: the number of samples between two consecutive analysis
        frames (``speed * synthesis_hop`` by default). If ``analysis_hop`` is
        set, the ``speed`` parameter will be ignored.
    :type analysis_hop: int, optional
    :param synthesis_hop: the number of samples between two consecutive
        synthesis frames (``frame_length // 2`` by default).
    :type synthesis_hop: int, optional
    :param tolerance: the maximum number of samples that the analysis frame can
        be shifted.
    :type tolerance: int
    :returns: a :class:`audiotsm.base.tsm.TSM` object
    """
    # pylint: disable=too-many-arguments
    if synthesis_hop is None:
        synthesis_hop = frame_length // 2

    if analysis_hop is None:
        analysis_hop = int(synthesis_hop * speed)

    if tolerance is None:
        tolerance = frame_length // 2

    analysis_window = None
    synthesis_window = hanning(frame_length)

    converter = WSOLAConverter(channels, frame_length, synthesis_hop,
                               tolerance)

    return AnalysisSynthesisTSM(
        converter, channels, frame_length, analysis_hop, synthesis_hop,
        analysis_window, synthesis_window, tolerance,
        tolerance + synthesis_hop)



# -*- coding: utf-8 -*-

"""
:class:`~audiotsm.base.tsm.TSM` objects use :class:`~audiotsm.io.base.Reader`
objects as input and :class:`~audiotsm.io.base.Writer` objects as output.

The :mod:`audiotsm.io` package provides Readers and Writers allowing to use
`numpy arrays`_ or `wav files`_ as input or output of a
:class:`~audiotsm.base.tsm.TSM`, to `play the output in real-time <play in
real-time>`_, as well as base classes to `implement your own <implementing your
own>`_ Readers and Writers.
"""



# -*- coding: utf-8 -*-

"""
The :mod:`audiotsm` module provides several time-scale modification procedures:

- :func:`~audiotsm.ola` (Overlap-Add), which should only be used for percussive
  audio signals;
- :func:`~audiotsm.wsola` (Waveform Similarity-based Overlap-Add), which should
  give good results on most inputs.

.. note::

    If you are not sure which procedure and parameters to use, using
    :func:`~audiotsm.wsola` with the default parameters should work in most
    cases.

Each of the function of this module returns a :class:`~audiotsm.base.tsm.TSM`
object which implements a time-scale modification procedure.

.. autofunction:: audiotsm.ola
.. autofunction:: audiotsm.wsola
"""

__version__ = "0.1.0"


from .ola import ola
from .wsola import wsola



# -*- coding: utf-8 -*-

"""
Tests for the audiotsm.utils.windows package.
"""

import pytest
import numpy as np
from numpy.testing import assert_almost_equal

from audiotsm.utils import windows


@pytest.mark.parametrize("buffer, window, out", [
    ([[]], None, [[]]),
    ([[]], [], [[]]),
    ([[1, 2, 3], [4, 5, 6]], None, [[1, 2, 3], [4, 5, 6]]),
    ([[1, 2, 3], [4, 5, 6]], [1, 1, 1], [[1, 2, 3], [4, 5, 6]]),
    ([[1, 2, 3], [4, 5, 6]], [2, 0.5, 1], [[2, 1, 3], [8, 2.5, 6]]),
])
def test_apply(buffer, window, out):
    """Run tests for apply."""
    buffer = np.array(buffer, dtype=np.float32)
    windows.apply(buffer, window)
    assert_almost_equal(buffer, np.array(out))


@pytest.mark.parametrize("length, window", [
    (0, []),
    (1, [0.]),
    (3, [0, 0.75, 0.75]),
    (4, [0, 0.5, 1, 0.5]),
    (8, [0, 0.14644661, 0.5, 0.85355339, 1, 0.85355339, 0.5, 0.146446611]),
])
def test_hanning(length, window):
    """Run tests for hanning."""
    assert_almost_equal(windows.hanning(length), np.array(window))


@pytest.mark.parametrize("window1, window2, out", [
    (None, None, None),
    (None, [], []),
    (None, [1, 2], [1, 2]),
    ([], None, []),
    ([1, 2], None, [1, 2]),
    ([1, 2], [3, 4], [3, 8]),
])
def test_product(window1, window2, out):
    """Run tests for product."""
    if window1 is not None:
        window1 = np.array(window1)
    if window2 is not None:
        window2 = np.array(window2)

    if out is None:
        assert windows.product(window1, window2) is None
    else:
        assert_almost_equal(windows.product(window1, window2), np.array(out))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# pylint: disable=invalid-name
"""
AudioTSM
~~~~~~~~

AudioTSM is a python library for real-time audio time-scale modification
procedures, i.e. algorithms that change the speed of an audio signal without
changing its pitch.

:copyright: (c) 2017 by Muges.
:license: MIT, see LICENSE for more details.
"""

import os
import re
import sys
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
try:
    from sphinx.setup_command import BuildDoc
except ImportError:
    BuildDoc = None


def find_version():
    """Read the package's version from __init__.py"""
    version_filename = os.path.abspath("audiotsm/__init__.py")
    with open(version_filename) as fileobj:
        version_content = fileobj.read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_content, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = ['--cov=audiotsm']
        self.test_suite = True

    def run_tests(self):
        import pytest

        errcode = pytest.main(self.test_args)
        sys.exit(errcode)


with open('README.rst', 'r') as f:
    long_description = f.read()


setup(
    name="audiotsm",
    version=find_version(),
    description="A real-time audio time-scale modification library",
    long_description=long_description,
    license="MIT",
    url="https://github.com/Muges/audiotsm",
    author="Muges",
    author_email="git@muges.fr",

    packages=find_packages(),

    install_requires=[
        "numpy",
    ],
    tests_require=[
        "pytest",
        "pytest-coverage",
        "sounddevice",
    ],
    extras_require={
        "doc": ["sphinx", "sphinx_rtd_theme"],
        "StreamWriter": ["sounddevice"]
    },

    cmdclass={
        'doc': BuildDoc,
        'test': PyTest,
    },

    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Multimedia :: Sound/Audio"
    ]
)



# -*- coding: utf-8 -*-

"""
The :mod:`audiotsm.io.array` module provides a Reader and Writers allowing to
use a :class:`numpy.ndarray` as input or output of a
:class:`~audiotsm.base.tsm.TSM` object.
"""

import numpy as np

from . import base


class ArrayReader(base.Reader):
    """A :class:`~audiotsm.io.base.Reader` allowing to use
    :class:`numpy.ndarray` as input of a :class:`~audiotsm.base.tsm.TSM`
    object.

    :param data: a matrix of shape (``m``, ``n``), with ``m`` the number of
        channels and ``n`` the length of the buffer, where the samples will be
        read.
    :type data: :class:`numpy.ndarray`
    """
    def __init__(self, data):
        self._data = data

    @property
    def channels(self):
        return self._data.shape[0]

    @property
    def empty(self):
        return self._data.shape[1] == 0

    def read(self, buffer):
        if buffer.shape[0] != self._data.shape[0]:
            raise ValueError("the two buffers should have the same number of "
                             "channels")

        # Number of samples to read
        n = min(buffer.shape[1], self._data.shape[1])
        np.copyto(buffer[:, :n], self._data[:, :n])

        # Remove the samples that were read
        self._data = self._data[:, n:]

        return n

    def skip(self, n):
        if n > self._data.shape[1]:
            n = self._data.shape[1]

        self._data = self._data[:, n:]

        return n


class ArrayWriter(base.Writer):
    """A :class:`~audiotsm.io.base.Writer` allowing to get the output of a
    :class:`~audiotsm.base.tsm.TSM` object as a :class:`numpy.ndarray`.

    Writing to an :class:`~audiotsm.io.array.ArrayWriter` will add the data at
    the end of the :attr:`~audiotsm.io.array.ArrayWriter.data` attribute.

    :param channels: the number of channels of the signal.
    :type channels: int
    """

    def __init__(self, channels):
        self._channels = channels
        self._data = []

    @property
    def channels(self):
        return self._channels

    def write(self, buffer):
        if buffer.shape[0] != self._channels:
            raise ValueError("the buffer should have the same number of "
                             "channels as the ArrayWriter")

        self._data.append(np.copy(buffer))

        return buffer.shape[1]

    @property
    def data(self):
        """A :class:`numpy.ndarray` of shape (``m``, ``n``), with ``m`` the
        number of channels and ``n`` the length of the data, where the samples
        have written."""
        if not self._data:
            return np.ndarray((self._channels, 0), dtype=np.float32)

        data = np.concatenate(self._data, axis=1)
        self._data = [data]

        return data


class FixedArrayWriter(base.Writer):
    """A :class:`~audiotsm.io.base.Writer` allowing to use
    :class:`numpy.ndarray` as output of a TSM object.

    Contrary to an :class:`~audiotsm.io.array.ArrayWriter`, a
    :class:`~audiotsm.io.array.FixedArrayWriter` takes the buffer in which the
    data will be written as a parameter of its constructor. The buffer is of
    fixed size, and it will not be possible to write more samples to the
    :class:`~audiotsm.io.array.FixedArrayWriter` than the buffer can contain.

    :param data: a matrix of shape (``m``, ``n``), with ``m`` the number of
        channels and ``n`` the length of the buffer, where the samples will be
        written.
    :type data: :class:`numpy.ndarray`
    """
    def __init__(self, data):
        self._data = data

    @property
    def channels(self):
        return self._data.shape[0]

    def write(self, buffer):
        if buffer.shape[0] != self._data.shape[0]:
            raise ValueError("the two buffers should have the same number of "
                             "channels")

        # Number of samples to write
        n = min(buffer.shape[1], self._data.shape[1])
        np.copyto(self._data[:, :n], buffer[:, :n])

        self._data = self._data[:, n:]

        return n



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
audiotsmcli
~~~~~~~~~~~

Change the speed of an audio file without changing its pitch.
"""

import argparse
import os

from audiotsm import ola
from audiotsm import wsola
from audiotsm.io.stream import StreamWriter
from audiotsm.io.wav import WavReader, WavWriter


def create_writer(output, reader):
    """Create a Writer with the same parameters as ``reader``.

    :param output: the path of the output wav file. If it is None, the output
        will be played with a StreamWriter.
    :type output: str
    :param reader: the WavReader used as input of the TSM
    """
    if output:
        return WavWriter(output, reader.channels, reader.samplerate)

    return StreamWriter(reader.channels, reader.samplerate)


def create_tsm(name, channels, parameters):
    """Create a TSM object given the method name and its parameters."""
    if name == "wsola":
        return wsola(channels, **parameters)
    if name == "ola":
        return ola(channels, **parameters)

    raise ValueError("unknown TSM method: {}".format(name))


def main():
    """Change the speed of an audio file without changing its pitch."""

    # Parse command line arguments
    parser = argparse.ArgumentParser(description=(
        "Change the speed of an audio file without changing its pitch."))
    parser.add_argument(
        '-s', '--speed', metavar="S", type=float, default=1.,
        help="Set the speed ratio (e.g 0.5 to play at half speed)")
    parser.add_argument(
        '-m', '--method', type=str, default="wsola",
        help="Select the TSM method (ola or wsola)")
    parser.add_argument(
        '-l', '--frame-length', metavar='N', type=int, default=None,
        help="Set the frame length to N.")
    parser.add_argument(
        '-a', '--analysis-hop', metavar='N', type=int, default=None,
        help="Set the analysis hop to N.")
    parser.add_argument(
        '--synthesis-hop', metavar='N', type=int, default=None,
        help="Set the synthesis hop to N.")
    parser.add_argument(
        '-t', '--tolerance', metavar='N', type=int, default=None,
        help="Set the tolerance to N (only used when method is set wsola).")
    parser.add_argument(
        '-o', '--output', metavar='FILENAME', type=str, default=None,
        help="Write the output in the wav file FILENAME.")
    parser.add_argument(
        'input_filename', metavar='INPUT_FILENAME', type=str,
        help="The audio input file")

    args = parser.parse_args()

    if not os.path.isfile(args.input_filename):
        parser.error(
            'The input file "{}" does not exist.'.format(args.input_filename))

    # Get TSM method parameters
    parameters = {}
    if args.speed:
        parameters['speed'] = args.speed
    if args.frame_length:
        parameters['frame_length'] = args.frame_length
    if args.analysis_hop:
        parameters['analysis_hop'] = args.analysis_hop
    if args.synthesis_hop:
        parameters['synthesis_hop'] = args.synthesis_hop
    if args.tolerance is not None and args.method == "wsola":
        parameters['tolerance'] = args.tolerance

    # Run the TSM procedure
    with WavReader(args.input_filename) as reader:
        with create_writer(args.output, reader) as writer:
            tsm = create_tsm(args.method, reader.channels, parameters)
            tsm.run(reader, writer)


if __name__ == "__main__":
    main()
